from ...misc import ArgRepr


# ToDo: Add strip only
class Lower(ArgRepr):
    """Lowercase strings, optionally stripping characters left and/or right.

    Parameters
    ----------
    lstrip: str, optional
        Characters to strip from the left of a string. Defaults to ``None``,
        resulting in only whitespaces to be stripped.
    rstrip: str, optional
        Characters to strip from the right of a string. Defaults to ``None``,
        resulting in only whitespaces to be stripped.

    """

    def __init__(
            self,
            lstrip: str | None = None,
            rstrip: str | None = None
    ) -> None:
        super().__init__(lstrip, rstrip)
        self.lstrip = lstrip
        self.rstrip = rstrip

    def __call__(self, arg: str) -> str:
        """Lowercase strings, optionally stripping left and/or right.

        Parameters
        ----------
        arg: str
            String to put into lowercase and optionally strip characters from.

        Returns
        -------
        str
            The lowercased and optionally stripped string.

        """
        return str(arg).lower().lstrip(self.lstrip).rstrip(self.rstrip)



from pathlib import Path


def resolve(path: str) -> str:
    """Resolve the given path to its full name using pathlib's Path.

    Parameters
    ----------
    path: str
        The path to resolve.

    Returns
    -------
    str
        The resolved path

    """
    return str(Path(path).resolve())



import warnings
from abc import ABC, abstractmethod
from typing import Any, TypedDict
from collections import OrderedDict
from pathlib import Path
import torch as pt
from ...misc import NotFound, LiteralNotFound
from ..types import Tensor, Module, Optimizer, LRScheduler

__all__ = [
    'State',
    'Checkpoint',
    'InMemory',
    'OnDisk'
]

class State(TypedDict):
    """Container for a snapshot of the state of model training."""
    epoch: int  #: The current epoch.
    loss: float  #: The (train or test) loss at the current epoch.
    model: OrderedDict[str, Any]  #: The state-dict of the model.
    optimizer: dict[str, Any]  #: The state-dict of the optimizer.
    scheduler: dict[str, Any]  #: The state-dict of the scheduler.


class Checkpoint(ABC):
    """Base class to inherit from when implementing custom checkpoints."""

    def __init__(self) -> None:
        self.__counter = 0

    @property
    def counter(self) -> int:
        """How many checkpoints were saved since the last reset."""
        return self.__counter

    def save(
            self,
            epoch: int,
            loss: float,
            model: Module,
            optimizer: Optimizer | None = None,
            scheduler: LRScheduler | None = None
    ) -> None:
        """User-facing method for saving a checkpoint during training.

        Parameters
        ----------
        epoch: int
            The current epoch.
        loss: float
            The (train or test) loss at the current `epoch`.
        model: Module
            The model to checkpoint the state of.
        optimizer: Optimizer, optional
            The optimizer to checkpoint the state of. Defaults to ``None``.
        scheduler: LRScheduler, optional
            The scheduler to checkpoint the state of. Defaults to ``None``.

        """
        self.__counter += 1
        self._save_state({
            'epoch': epoch,
            'loss': loss,
            'model': model.state_dict(),
            'optimizer': {} if optimizer is None else optimizer.state_dict(),
            'scheduler': {} if scheduler is None else scheduler.state_dict()
        })

    def load(
            self,
            model: Module,
            optimizer: Optimizer | None = None,
            scheduler: LRScheduler | None = None
    ) -> tuple[int, float]:
        """User-facing method for loading a checkpoint during training.

        The state retrieved from the checkpoint is merged into the state of
        the objects to update, such that loading a checkpoint before one has
        been saved results in unchanged objects.

        Parameters
        ----------
        model: Module
            The model to load the state for in-place.
        optimizer: Optimizer, optional
            The optimizer to load the state for in-place. Defaults to ``None``.
        scheduler: LRScheduler, optional
            The scheduler to load the state for in-place. Defaults to ``None``.

        Returns
        -------
        epoch: int
            The epoch of the checkpoint.
        loss: float
            The loss from the checkpoint.

        """
        buffer = self._load_state()
        model.load_state_dict(model.state_dict() | buffer['model'])
        if optimizer is not None:
            optimizer.load_state_dict(
                optimizer.state_dict() | buffer['optimizer']
            )
        if scheduler is not None:
            scheduler.load_state_dict(
                scheduler.state_dict() | buffer['scheduler']
            )
        return buffer['epoch'], buffer['loss']

    def reset_parameters(self) -> None:
        """Hard-resets the checkpoint into a pristine, initial state."""
        self.__counter = 0
        self._save_state({
            'epoch': 0,
            'loss': float('inf'),
            'model': OrderedDict(),
            'optimizer': {},
            'scheduler': {}
        })

    @abstractmethod
    def _save_state(self, state: State) -> None:
        """Subclasses must implement how exactly the state is persisted.

        Parameters
        ----------
        state: State
            The state to persist.

        Warnings
        --------
        Make sure the persisted state is decoupled from the actual
        state (e.g., copied) so that it does not change as model
        training continues!

        """
        ...

    @abstractmethod
    def _load_state(self) -> State:
        """Subclasses must implement how exactly the state is retrieved.

        Returns
        -------
        State

        Warnings
        --------
        Make sure that the returned state is decoupled from the persisted
        state (e.g., copied) so that the latter cannot change as model
        training continues!

        """
        ...


class InMemory(Checkpoint):
    """Checkpoint in CPU memory regardless of the device training runs on."""

    def __init__(self) -> None:
        super().__init__()
        self.state = {
            'epoch': 0,
            'loss': float('inf'),
            'model': OrderedDict(),
            'optimizer': {},
            'scheduler': {}
        }

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'

    def _to_cpu(self, obj: Any) -> Any:
        """Recursively copy tensors to CPU and remember original devices."""
        if isinstance(obj, dict):
            clone = {}
            for key, value in obj.items():
                clone[key] = self._to_cpu(value)
            return clone
        if isinstance(obj, list):
            return [self._to_cpu(item) for item in obj]
        if isinstance(obj, Tensor):
            clone = pt.empty_like(obj, device='cpu').copy_(obj, True)
            clone.target = obj.device
            return clone
        return obj

    def _to_device(self, obj: Any) -> Any:
        """Recursively copy backed-up tensors to their original devices."""
        if isinstance(obj, dict):
            clone = {}
            for key, value in obj.items():
                clone[key] = self._to_device(value)
            return clone
        if isinstance(obj, list):
            return [self._to_device(item) for item in obj]
        if isinstance(obj, pt.Tensor):
            return pt.empty_like(obj, device=obj.target).copy_(obj, True)
        return obj

    def _save_state(self, state: State) -> None:
        """Create a deep copy of the combined state in CPU memory."""
        self.state = self._to_cpu(state)

    def _load_state(self) -> State:
        """Create a deep copy of the saved state on the original device(s)."""
        return self._to_device(self.state)


class OnDisk(Checkpoint):
    """Checkpoint the current state of training on disk.

    Parameters
    ----------
    path: str
        Full path to and name of the file used to persist state.
    create: bool, optional
        What to do if the directory where the checkpoint should be saved does
        not exist. Defaults to ``False``.
    not_found: str, optional
        What to do if a checkpoint is loaded from the specified `path`
        but none is there yet. One of "ignore", "warn", or "raise".
        Defaults to "raise". Use the ``NotFound`` enum to avoid typos!

    See Also
    --------
    ~swak.misc.NotFound

    """

    def __init__(
            self,
            path: str,
            create: bool = False,
            not_found: NotFound | LiteralNotFound = NotFound.WARN
    ) -> None:
        super().__init__()
        self.path = str(Path(str(path).strip()).resolve())
        self.create = create
        self.not_found = str(not_found).strip().lower()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.path}', {self.create})"

    def _save_state(self, state: State) -> None:
        """Save the combined state to file."""
        if self.create:
            Path(self.path).parent.mkdir(parents=True, exist_ok=True)
        pt.save(state, self.path)

    def _load_state(self) -> State:
        """Retrieve the combined state from file."""
        try:
            return pt.load(self.path, weights_only=True)
        except (FileNotFoundError, EOFError) as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'Checkpoint {} not found!\nCreating an empty one.'
                    warnings.warn(msg.format(self.path))
                    self.reset_parameters()
                case NotFound.IGNORE:
                    self.reset_parameters()
                case _:
                    raise error
        return pt.load(self.path, weights_only=True)



"""Tools to interact with hosted cloud services.

Current supports elements of the Google Cloud Project (GCP) and Amazon Web
Services (AWS).

"""

from importlib.util import find_spec

required = (
    'google_cloud_bigquery',
    'google_cloud_storage',
    'pandas_gbq',
    'boto3'
)

if any(find_spec(package) for package in required) is None:
    msg = 'Install {} with the [cloud] extra to unlock this subpackage!'
    raise ImportError(msg.format(__package__.split('.')[0]))



from typing import Any, Self
from collections.abc import Callable, Iterator
from abc import ABC, abstractmethod
from pandas import DataFrame
from torch import Tensor, dtype, device
from torch.nn import Module, Dropout, AlphaDropout
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

type Functional = Callable[[Tensor], Tensor]
type Tensors1T = tuple[Tensor]
type Tensors2T = tuple[Tensor, Tensor]
type Tensors3T = tuple[Tensor, Tensor, Tensor]
type Tensors4T = tuple[Tensor, Tensor, Tensor, Tensor]
type Tensors = tuple[Tensor, ...]
type Dtype = dtype
type Device = device
type Drop = Dropout | AlphaDropout
type Batch = tuple[tuple[Any, ...], Tensor]
type Batches = Iterator[Batch]

__all__ = [
    'DataFrame',
    'Tensor',
    'Module',
    'Functional',
    'Tensors1T',
    'Tensors2T',
    'Tensors3T',
    'Tensors4T',
    'Tensors',
    'Dtype',
    'Device',
    'Drop',
    'Optimizer',
    'LRScheduler',
    'Batch',
    'Batches',
    'Resettable',
    'Block'
]


class Resettable(Module, ABC):
    """Abstract base class for Modules with a ``reset_parameters`` method."""

    def __init__(self, *_: Any, **__: Any) -> None:
        super().__init__()

    @abstractmethod
    def reset_parameters(self) -> None:
        """Subclasses implement in-place reset of all internal parameters."""
        ...


class Block(Resettable):
    """Abstract base class for stackable/repeatable neural-network components.

    The input and output tensors of such components must have the same
    dimensions and sizes!

    """

    @abstractmethod
    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        ...



from typing import Any
from io import BytesIO
from boto3.s3.transfer import TransferConfig
from pandas import DataFrame as PandasFrame
from polars import DataFrame as PolarsFrame
from ...misc import ArgRepr
from .s3 import S3

type Frame = PandasFrame | PolarsFrame


class DataFrame2S3Parquet(ArgRepr):
    """Upload a pandas or polars dataframe to an S3 bucket.

    Parameters
    ----------
    s3: S3
        An instance of a wrapped S3 client.
    bucket: str
        The name of the bucket to upload to.
    prefix: str, optional
        The prefix of the parquet file to upload the dataframe to. May include
        any number of string placeholders (i.e., pairs of curly brackets) that
        will be interpolated when instances are called.
        Defaults to an empty string.
    extra_kws: dict, optional
        Passed on as ``ExtraArgs`` to the `upload_fileobj <meth_>`__ method of
        the client. See the `docs <doc_>`__ for all options.
    upload_kws: dict, optional
        Passed on as ``Config`` to the `upload_fileobj <meth_>`__ method of
        the client. See the `docs <doc_>`__ for all options.
    **kwargs
        Additional keyword arguments are passed on to the ``to_parquet`` method
        of the dataframe.


    .. _meth: https://boto3.amazonaws.com/v1/documentation/api/latest/reference
             /services/s3/client/upload_fileobj.html
    .. _doc: https://boto3.amazonaws.com/v1/documentation/api/latest/
            reference/customizations/s3.html#boto3.s3.transfer.TransferConfig

    See Also
    --------
    S3

    """

    def __init__(
            self,
            s3: S3,
            bucket: str,
            prefix: str = '',
            extra_kws: dict[str, Any] | None = None,
            upload_kws: dict[str, Any] | None = None,
            **kwargs: Any
    ) -> None:
        self.s3 = s3
        self.bucket = self.__strip(bucket)
        self.prefix = self.__strip(prefix)
        self.extra_kws = {} if extra_kws is None else extra_kws
        self.upload_kws = {} if upload_kws is None else upload_kws
        self.kwargs = kwargs
        super().__init__(
            self.s3,
            self.bucket,
            self.prefix,
            extra_kws=extra_kws,
            upload_kws=upload_kws,
            **self.kwargs
        )

    @staticmethod
    def __strip(attr: str | None) -> str:
        """Strip leading and trailing whitespaces from string arguments."""
        return attr if attr is None else attr.strip(' /')

    def __call__(self, df: Frame, *parts: str) -> tuple[()]:
        """Write a pandas or polars dataframe to S3 object storage.

        Parameters
        ----------
        df: DataFrame
            The pandas or polars dataframe to upload.
        *parts: str
            Fragments that will be interpolated into the `prefix` given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `prefix`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        key = self.prefix.format(*parts).strip()
        with BytesIO() as buffer:
            df.to_parquet(buffer, **self.kwargs)
            buffer.seek(0)
            self.s3.client.upload_fileobj(
                Fileobj=buffer,
                Bucket=self.bucket,
                Key=key,
                ExtraArgs=self.extra_kws,
                Config=TransferConfig(**self.upload_kws)
            )
        return ()



"""Convenient classes (and functions) that do not fit any other category."""

from typing import Any, Self, overload
from collections.abc import Iterable
from functools import cached_property, singledispatchmethod
from itertools import chain
import torch as pt
import torch.nn as ptn
from ..misc import ArgRepr
from .types import (
    Tensor,
    Module,
    Tensors,
    Tensors2T,
    Dtype,
    Device,
    Block,
    Resettable
)
from .exceptions import (
    CompileError,
    ShapeError,
    DeviceError,
    DTypeError,
    ValidationErrors
)

__all__ = [
    'identity',
    'Identity',
    'Finalizer',
    'NegativeBinomialFinalizer',
    'Compile',
    'Cat',
    'LazyCatDim0'
]


def identity(tensor: Tensor, **_: Any) -> Tensor:
    """Simply pass through the argument and ignore keyword arguments.

    This is a placeholder for instances where a default function is required.

    Parameters
    ----------
    tensor: Tensor
        Any argument (typically a tensor) to be passed straight through.

    Returns
    -------
    Tensor
        The tensor passed in as argument.

    """
    return tensor


class Identity(Block):
    """PyTorch module that passes a tensor right through, doing nothing.

    This is a placeholder for instances where a default module is required.
    Providing any number of (keyword) arguments on instantiation is permitted,
    but they will be ignored.

    """

    def __init__(self, *_, **__) -> None:
        super().__init__()

    def forward(self, tensor: Tensor, **_: Any) -> Tensor:
        """Simply pass through the argument and ignore keyword arguments.

        Parameters
        ----------
        tensor: Tensor
            Any argument (typically a tensor) to be passed straight through.

        Returns
        -------
        Tensor
            The tensor passed in as argument.

        """
        return tensor

    def reset_parameters(self) -> None:
        """Does nothing because there are no internal parameters to reset."""

    def new(self, *_, **__) -> Self:
        """Return a fresh, new instance.

        Providing any number of (keyword) arguments is permitted, but they will
        be ignored.

        Returns
        -------
        Identity
            A fresh, new instance of itself.

        """
        return self.__class__()


class Finalizer(Resettable):
    """Extract one or more numbers from the final layer of a neural network.

    Instances of this class serve as a convenient final layer in any neural
    network, no matter if it is for regression, for classification, for
    multiple targets, or if you predict the parameters of a probability
    distribution. The last activations of your network are going to be passed
    through as many linear layers as you need outputs, each passed through
    their own (and potentially different) non-linearity to give you the desired
    number of outputs and the desired value range for each output.

    Parameters
    ----------
    mod_dim: int
        The size of the last dimension of the input tensor, essentially the
        "width" of the neural network before it is to be collapsed to the
        final output.
    *activations: Module
        Specify as many activations (instances of PyTorch ``Module``) as you
        want outputs (e.g., ``Sigmoid()`` for binary classification,
        ``Softplus()`` for strictly positive regression targets, etc.). For
        unbounded regression targets, where you want no activation function at
        all, use ``Identity()``.
    **kwargs
        Keyword arguments are passed on to all linear layers.

    See Also
    --------
    Identity

    """

    def __init__(
            self,
            mod_dim: int,
            *activations: Module,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.activations: tuple[Module, ...] = activations
        self.kwargs = kwargs
        self.finalize = ptn.ModuleList(
            ptn.Sequential(ptn.Linear(mod_dim, 1, **kwargs), activate)
            for activate in activations
        )

    @property
    def n_out(self) -> int:
        """Number of final outputs."""
        return len(self.activations)

    def forward(self, inp: Tensor) -> Tensors:
        """Forward pass for extracting outputs from the final hidden layer.

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        tuple
            As many tensors are returned as there were `activations` specified.
            Each tensor has the same dimension and size, viz., that of the
            `inp` with the size of the last dimension shrunk to 1.

        """
        return tuple(finalize(inp) for finalize in self.finalize)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        for finalize in self.finalize:
            finalize[0].reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            *activations: Module,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            The size of the last dimension of the input tensor. Overwrites the
            `mod_dim` of the current instance if given. Defaults to ``None``.
        *activations: Module
            Activation functions replace the ones in the current instance if
            any are given. If none are given, the new instance will have the
            same as the present instance.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layers together.

        Returns
        -------
        Finalizer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            *(activations or self.activations),
            **(self.kwargs | kwargs)
        )


class NegativeBinomialFinalizer(Resettable):
    """Consistent mean and standard deviation for over-dispersed counts.

    When regressing potentially over-dispersed counts data, you might want to
    use the negative log-likelihood as loss function. However, this is only
    defined if the variance is greater than the mean. Following D. Salinas
    `et al.`, [1]_ this can be achieved by extracting a (positive) mean value
    and a (positive) scale factor from the last hidden layer of your network,
    and letting the variance be the sum of mean and the scaled square of the
    mean.

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension.
    beta: float, optional
        Scaling parameter ot the `Softplus <https://pytorch.org/docs/stable/
        generated/torch.nn.Softplus.html#torch.nn.Softplus>`__ activation
        function. Defaults to 1.
    threshold: float, optional
        The `Softplus <https://pytorch.org/docs/stable/generated/
        torch.nn.Softplus.html#torch.nn.Softplus>`__ activation is approximated
        as a linear function for values greater than this. Defaults to 20.
    **kwargs
        Keyword arguments are passed on to the linear layers.

    See Also
    --------
    swak.pt.losses.NegativeBinomialLoss
    swak.pt.dists.MuSigmaNegativeBinomial

    References
    ----------
    .. [1] D. Salinas, V. Flunkert, and J. Gasthaus, and T. Pfister, `DeepAR:
           Probabilistic Forecasting with Autoregressive Recurrent Networks`,
           `arXiv:1704.04110v3 <https://arxiv.org/pdf/1704.04110>`__ (2019).

    """

    def __init__(
            self,
            mod_dim: int,
            beta: float = 1.0,
            threshold: float = 20.0,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.beta = beta
        self.threshold = threshold
        self.kwargs = kwargs
        self.mu = ptn.Linear(mod_dim, 1, **kwargs)
        self.alpha = ptn.Linear(mod_dim, 1, **kwargs)
        self.activate = ptn.Softplus(beta, threshold)

    def forward(self, inp: Tensor) -> Tensors2T:
        """Forward pass for generating mean and matching standard deviation.

        Parameters
        ----------
        inp: Tensor
            The activations after the last hidden layer in your network.
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        tuple
            A tensor with the predicted mean values and a tensor with the
            predicted standard deviations that are guaranteed to be greater
            or equal to the square root of the mean. Both have the same
            dimension and size, viz., that of the `inp` with the size of the
            last dimension shrunk to 1.

        """
        mu = self.activate(self.mu(inp))
        alpha = self.activate(self.alpha(inp))
        return mu, (mu * (1.0 + mu * alpha)).sqrt()

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.mu.reset_parameters()
        self.alpha.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            beta: float | None = None,
            threshold: float | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            The size of the last dimension of the input tensor. Overwrites the
            `mod_dim` of the current instance if given. Defaults to ``None``.
        beta: float, optional
            Scaling parameter ot the ``Softplus`` activation function.
            Overwrites the `beta` of the current instance if given.
            Defaults to ``None``.
        threshold: float, optional
            The ``Softplus`` activation is approximated as a linear function
            for values greater than this. Overwrites the `threshold` of the
            current instance if given. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layers together.

        Returns
        -------
        NegativeBinomialFinalizer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.beta if beta is None else beta,
            self.threshold if threshold is None else threshold,
            **(self.kwargs | kwargs)
        )


class Compile:
    """Partial of the ``compile`` top-level function or Module method.

    Parameters
    ----------
    inplace: bool, optional
        Whether to compile the model in place (by calling its ``compile``
        method) or create a new, compiled instance. Defaults to ``True``
    model: Module, optional
        For convenience, the model to compile can already be given at
        instantiation. However, Nothing will happen until instances are called.
    **kwargs
        Additional keyword arguments are forwarded to the ``compile`` function
        or method call. See the `Documentation <https://pytorch.org/docs/
        stable/generated/torch.compile.html#torch-compile>`__ for details.

    """

    def __init__(
            self,
            inplace: bool = True,
            model: Module | None = None,
            **kwargs: Any
    ) -> None:
        self.inplace = inplace
        self.model = model
        self.kwargs = kwargs

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        model = None if self.model is None else 'model'
        kwargs = (f'{k}={str(v)}' for k, v in self.kwargs.items())
        kwargs = ', '.join(kwargs)
        sep = ', ' if kwargs else ''
        return f'{cls}({self.inplace}, {model}{sep}{kwargs})'

    def __call__(self, model: Module | None = None, **kwargs: Any) -> Module:
        """Compile a Module with the given options.

        Parameters
        ----------
        model: Module, optional
            If no model was given on instantiation, one must be given here.
            Otherwise, there would be nothing to compile. If a model was given
            on instantiation and one is given here, the latter replaces the
            former. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into those given at
            instantiation and then forward to the ``compile`` function
            or method call. See the `Documentation <https://pytorch.org/docs/
            stable/generated/torch.compile.html#torch-compile>`__ for details.

        Returns
        -------
        Module
            The compiled module.

        Raises
        ------
        CompileError
            If no model was given, neither at instantiation, nor when calling
            instances.

        """
        model = self.model if model is None else model
        if model is None:
            raise CompileError('No model to compile!')
        merged_kwargs = self.kwargs | kwargs
        if self.inplace:
            model.compile(**merged_kwargs)
            return model
        return pt.compile(model, **merged_kwargs)


class Cat(ArgRepr):
    """Simple partial of PyTorch's top-level `cat` function.

    Parameters
    ----------
    dim: int, optional
        The dimension along which to concatenate the tensors. Defaults to 0.

    """

    def __init__(self, dim: int = 0) -> None:
        super().__init__(dim)
        self.dim = dim

    def __call__(self, tensors: Tensors | list[Tensor]) -> Tensor:
        """Concatenate the given tensors along one of their dimensions.

        Parameters
        ----------
        tensors: tuple or list of tensors
            Tensors to concatenate along an existing dimension.

        Returns
        -------
        Tensor
            The concatenated tensors.

        """
        return pt.cat(tensors, dim=self.dim)


class LazyCatDim0:
    """Lazily concatenate a sequence of tensors along their first dimension.

    Concatenating a large number of even small tensors (or a small number of
    large tensors) causes a memory spike because, temporarily, two copies of
    all tensors are needed. Sometimes, this simply cannot be avoided.
    However, when only a small part of the full concatenation of all tensors is
    needed at any given time, e.g., when chopping off micro-batches of training
    data to feed to a model, the present class provides an alternative:
    Constituent tensors are kept as is and concatenation is only performed when
    slices or element(s) along the first dimension are requested. These are
    selected from the constituents first and only then concatenated (and,
    thus, copied). Slicing and element selection of further dimensions
    is delayed until selection and concatenation along the first dimension is
    completed.

    Parameters
    ----------
    tensors: iterable
        The iterable of tensors to cache, all of which must have the same
        number of dimensions and the exact same sizes in all dimensions but
        the first.

    Raises
    ------
    ShapeError
        If any tensor is a scalar, that is, has zero dimensions, or if the
        shape after the first dimension is not the same across all tensors.
        Also, if there are no tensors to wrap.
    DeviceError
        If tensors are spread over multiple devices.
    DTypeError
        If tensors have multiple dtypes.

    """

    def __init__(self, tensors: Iterable[Tensor]) -> None:
        self.__tensors = self._valid(tuple(tensors))

    @staticmethod
    def _valid(tensors: Tensors) -> Tensors:
        """Run a few validations on the homogeneity of the wrapped tensors."""
        if not tensors:
            msg = 'Expected a non-empty iterable of tensors!'
            raise ShapeError(msg)
        errors = []
        if any(tensor.dim() == 0 for tensor in tensors):
            msg = 'Scalar tensors can not be concatenated!'
            errors.append(ShapeError(msg))
        _, *shape = tensors[0].shape
        if any(list(tensor.shape[1:]) != shape for tensor in tensors[1:]):
            msg = 'All tensors must be of shape {} after the first dimension!'
            errors.append(ShapeError(msg.format(shape)))
        if any(tensor.device != tensors[0].device for tensor in tensors[1:]):
            msg = 'All tensors must be on the same device!'
            errors.append(DeviceError(msg))
        if any(tensor.dtype is not tensors[0].dtype for tensor in tensors[1:]):
            msg = 'All tensors must have the same dtype!'
            errors.append(DTypeError(msg))
        if errors:
            raise ValidationErrors('Validation failed', errors)
        return tensors

    @cached_property
    def lookup(self) -> tuple[tuple[int, int], ...]:
        """A lazily computed and cached lookup table for indices."""
        return tuple(
            (i, j)
            for i, tensor in enumerate(self.__tensors)
            for j in range(tensor.size(0))
        )

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        return f'{cls}(n={len(self.lookup)})'

    def __len__(self) -> int:
        return len(self.lookup)

    def __iter__(self) -> chain[Tensor]:
        return chain.from_iterable(self.__tensors)

    def __contains__(self, elem: Any) -> bool:
        return any(elem in tensor for tensor in self.__tensors)

    @singledispatchmethod
    def __getitem__(self, index) -> Tensor:
        idx, elem = self.lookup[index]
        return self.__tensors[idx][elem]

    @__getitem__.register
    def _(self, index: slice) -> Tensor:
        return pt.cat([
            self.__tensors[idx][elem:elem + 1]
            for idx, elem
            in self.lookup[index.start:index.stop]
        ])[::index.step]

    @__getitem__.register
    def _(self, index: list) -> Tensor:
        return pt.cat([
            self.__tensors[idx][elem:elem + 1]
            for idx, elem
            in (self.lookup[idx] for idx in index)
        ])

    @__getitem__.register
    def _(self, index: tuple) -> Tensor:
        if not index:
            return self[:]
        idx, *indices = index
        match idx:
            case slice():
                return self[idx][:, *indices]
            case int():
                return self[idx][*indices]


    @property
    def dtype(self) -> Dtype:
        """The common dtype of all cached tensors."""
        return self.__tensors[0].dtype

    @property
    def device(self) -> Device:
        """The common device of all cached tensors."""
        return self.__tensors[0].device

    @cached_property
    def shape(self) -> pt.Size:
        """The shape of the full concatenation of the wrapped tensors."""
        _, *shape = self.__tensors[0].shape
        return pt.Size([len(self), *shape])

    @overload
    def size(self, dim: int) -> int:
        ...

    @overload
    def size(self, dim: None = None) -> pt.Size:
        ...

    def size(self, dim = None):
        """The size of the full concatenation of the wrapped tensors.

        Parameters
        ----------
        dim: int, optional
            The dimension for which to return the size. Defaults to ``None``.
            If not given the `shape` of the full concatenation of all cached
            tensors is returned.

        Returns
        -------
        int
            If the size along a certain dimension was requested.
        Size
            A PyTorch ``Size`` object specifying the `shape` of the full
            concatenation of all cached tensors.

        """
        return self.shape[dim] if dim is not None else self.shape

    def to(self, *args: Any) -> Self:
        """Returns a new instance of self, wrapping transformed tensors.

        See the PyTorch `documentation <https://pytorch.org/docs/stable/
        generated/torch.Tensor.to.html#torch-tensor-to>`_ for possible call
        signatures, but keep in mind that all listed operations will create
        a copy of the wrapped tensors after all.

        """
        return self.__class__(tensor.to(*args) for tensor in self.__tensors)



from typing import Any
from collections.abc import Hashable, Iterable, Callable
from functools import singledispatchmethod
from numpy import dtype, ndarray
from pandas.core.dtypes.base import ExtensionDtype
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy
from pandas import DataFrame, Series
from ..misc.repr import ReprName
from ..misc import ArgRepr

type Type = str | type | dtype | ExtensionDtype
type Types = dict[Hashable, Type]
type Mask = list[bool] | Series | ndarray
type Condition = Callable[[DataFrame], Mask]
type Transform = dict[Hashable, Any] | Series | Callable[[Any], Any]
type Others = Series | DataFrame | list[Series | DataFrame]


class AsType(ReprName):
    """Partial of the ``pandas.DataFrame.astype`` method.

    Parameters
    ----------
    types: type or dict
        Single type or dictionary of column names and types, specifying type
        conversion of entire DataFrame or specific columns, respectively
    **kwargs
        Keyword arguments are passed on to the ``astype`` method call after
        the `types` argument.

    """
    def __init__(self, types: Type | Types, **kwargs: Any) -> None:
        super().__init__()
        self.types = types
        self.kwargs = kwargs

    def __repr__(self) -> str:
        kwargs = (f'{k}={self._repr(v)}' for k, v in self.kwargs.items())
        kwargs = ', '.join(kwargs)
        try:
            args = (f'{k!r}: {self._name(self.types[k])}' for k in self.types)
            args = '{' + ', '.join(args) + '}'
        except TypeError:
            args = self._name(self.types)
        signature = ', '.join(filter(None, [args, kwargs]))
        return f'{self.__class__.__name__}({signature})'

    def __call__(self, df: DataFrame) -> DataFrame:
        """Cast entire dataframe or columns thereof to specified types

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to type-cast.

        Returns
        -------
        DataFrame
            Pandas dataframe cast to new type or with columns cast to
            new types.

        """
        return df.astype(self.types, **self.kwargs)


class ColumnSelector(ArgRepr):
    """Select a single column of a (grouped) pandas dataframe as a series.

    This is simply a partial for calling a (grouped) dataframe's
    ``__getitem__`` method with a single argument (using the square-brackets
    accessor).

    Parameters
    ----------
    col: hashable
        Single DataFrame column to select.

    """

    def __init__(self, col: Hashable) -> None:
        self.col = self.__valid(col)
        super().__init__(col)

    @singledispatchmethod
    def __call__(self, df: DataFrame) -> Series:
        """Select a single column of a (grouped) pandas dataframe as series.

        Parameters
        ----------
        df: DataFrame or DataFrameGroupBy
            Pandas dataframe or grouped dataframe to select column from.

        Returns
        -------
        Series or SeriesGroupBy
            The selected column from the (grouped) dataframe.

        """
        return df[self.col]

    @__call__.register
    def _(self, grouped_df: DataFrameGroupBy) -> SeriesGroupBy:
        return grouped_df[self.col]

    @staticmethod
    def __valid(col: Hashable) -> Hashable:
        _ = hash(col)
        return col


class ColumnsSelector(ArgRepr):
    """Select one or more columns of a (grouped) pandas dataframe as dataframe.

    This is simply a partial for calling a (grouped) dataframe's
    ``__getitem__``  method with a list of arguments (using the square-brackets
    accessor).

    Parameters
    ----------
    col: Hashable, optional
        Column name or iterable thereof. Defaults to an empty tuple.
    *cols: Hashable
        Additional columns names.

    """

    def __init__(
            self,
            col: Hashable | Iterable[Hashable] = (),
            *cols: Hashable
    ) -> None:
        col = self.__valid(col)
        self.cols: tuple[Hashable, ...] = col + self.__valid(cols)
        super().__init__(*self.cols)

    @singledispatchmethod
    def __call__(self, df: DataFrame | DataFrameGroupBy) -> DataFrame:
        """Select the specified column(s) from a (grouped) pandas dataframe.

        Parameters
        ----------
        df: DataFrame or DataFrameGroupBy
            Pandas dataframe or grouped dataframe to select column(s) from.

        Returns
        -------
        DataFrame or DataFrameGroupBy
            The selected column(s) of the (grouped) dataframe.

        """
        return df[list(self.cols)]

    @__call__.register
    def _(self, grouped_df: DataFrameGroupBy) -> DataFrameGroupBy:
        return grouped_df[list(self.cols)]

    @staticmethod
    def __valid(cols: Hashable | Iterable[Hashable]) -> tuple[Hashable, ...]:
        """Ensure that the columns are indeed an iterable of hashables."""
        if isinstance(cols, str):
            return cols,
        try:
            _ = [hash(col) for col in cols]
        except TypeError:
            _ = hash(cols)
            return cols,
        return tuple(cols)


class ColumnMapper(ArgRepr):
    """Transform one column of a pandas dataframe into another.

    This is simply a partial of calling the ``map`` method on one column of a
    dataframe and assigning the result to the same or another column of the
    same dataframe.

    Parameters
    ----------
    src_col: Hashable
        Column to call the ``map`` method on.
    transform: callable, Mapping, or Series
        Function or mapping in the form of a dictionary or a pandas series.
    tgt_col: Hashable, optional
        Dataframe column to store the series resulting from the
        transformation. Defaults to `src_col`, thus overwriting it in place.
    **kwargs
        Keyword arguments are passed on to the call of the Series'
        ``map`` method.

    """

    def __init__(
            self,
            src_col: Hashable,
            transform: Transform,
            tgt_col: Hashable | None = None,
            **kwargs: Any
    ) -> None:
        self.src_col = src_col
        self.tgt_col = src_col if tgt_col is None else tgt_col
        self.transform = transform
        self.kwargs = kwargs
        name = transform if callable(transform) else type(transform)
        super().__init__(src_col, name, tgt_col, **kwargs)

    def __call__(self, df: DataFrame) -> DataFrame:
        """Called the ``map`` method on a specified column of a DataFrame.

        Cached keyword arguments are forwarded to the method call and
        the result is stored in the specified column of the DataFrame.

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe with the column to call the ``map`` method on.

        Returns
        -------
        DataFrame
            Pandas dataframe with the result of the column transformation in
            the specified column.


        """
        df[self.tgt_col] = df[self.src_col].map(self.transform, **self.kwargs)
        return df


class RowsSelector(ArgRepr):
    """Select rows from a pandas dataframe with a boolean mask or function.

    This is simply a partial for calling a dataframe's ``__getitem__``
    method (using the square-brackets accessor) with a callable that takes
    the dataframe as input, and produces a 1-D, boolean array-like structure
    (of the same length as the dataframe to select from).

    Parameters
    ----------
    condition: callable or array-like
        A callable that accepts a dataframe and produces a 1-D, boolean array-
        like structure of the same length

    """

    def __init__(self, condition: Condition) -> None:
        super().__init__(condition)
        self.condition = condition

    def __call__(self, df: DataFrame) -> DataFrame:
        """Select rows from a dataframe with the specified mask or condition.

        Parameters
        ----------
        df: DataFrame
            The pandas dataframe to select rows from.

        Returns
        -------
        DataFrame
            The pandas dataframe with only the selected rows.

        """
        return df[self.condition]


class FrameGroupBy(ArgRepr):
    """Simple partial of a pandas dataframe's ``groupby`` method.

    Parameters
    ----------
    *args
        Arguments to pass on to the ``groupby`` method call.
    **kwargs
        Keyword arguments to pass on to the ``groupby`` method call.

    Note
    ----
    For a full list of (keyword) arguments and their description, see the
    pandas `documentation <https://pandas.pydata.org/pandas-docs/stable/
    reference/api/pandas.DataFrame.groupby.html>`_.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, df: DataFrame) -> DataFrameGroupBy:
        """Call the dataframe's ``groupby`` method with the cached (kw)args.

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to group.

        Returns
        -------
        DataFrameGroupBy
            The grouped dataframe.

        """
        return df.groupby(*self.args, **self.kwargs)


class Join(ArgRepr):
    """Light wrapper around the pandas dataframe ``join`` method.

    Parameters
    ----------
    *args
        Arguments to pass on to the ``join`` method call.
    **kwargs
        Keyword arguments to pass on to the ``join`` method call.

    Note
    ----
    For a full list of (keyword) arguments and their description, see the
    pandas `join documentation <https://pandas.pydata.org/pandas-docs/stable/
    reference/api/pandas.DataFrame.join.html>`_.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, df: DataFrame, other: Others) -> DataFrame:
        """Join a dataframe with other dataframe(s) and/or series.

        Parameters
        ----------
        df: DataFrame
            Source dataframe on which the ``join`` method will be called.
        other: DataFrame, Series, or a list of any combination
            Index should be similar to one (or more) columns in `df`. If a
            series is passed, its name attribute must be set, and that will be
            used as the column name in the resulting joined dataframe.

        Returns
        -------
        DataFrame
            The joined dataframe.

        """
        return df.join(other, *self.args, **self.kwargs)


class Assign(ArgRepr):
    """Light wrapper around a pandas dataframe's ``assign`` method.

    Parameters
    ----------
    col: dict
        A dictionary with the names of newly created (or overwritten) columns
        as keys. If the values are callable, they are computed on the entire
        dataframe and assigned to the new columns. The callable must not
        the change input dataframe (though pandas doesn’t check it). If the
        values are not callable, e.g., a series, scalar, or array, they are
        simply assigned.
    **cols
        As in the original, the keyword arguments themselves serve as the
        name(s) of the new (or overwritten) column(s) and their values are set
        in the same way.

    """

    def __init__(self, col: dict[str, Any] | None = None, **cols: Any) -> None:
        self.cols = cols if col is None else col | cols
        super().__init__(**self.cols)

    def __call__(self, df: DataFrame) -> DataFrame:
        """Add new columns to a dataframe by calling its ``assign`` method.

        Parameters
        ----------
        df: DataFrame
            The dataframe to add new columns to.

        Returns
        -------
        DataFrame
            The input dataframe with new columns added.

        """
        return df.assign(**self.cols)



import math
from collections.abc import Iterable
import torch as pt
from torch.optim import AdamW
from torch.nn.utils import clip_grad_norm_
from tqdm import tqdm
from ...misc import ArgRepr
from ...funcflow import Curry
from ..types import Module, Optimizer, LRScheduler, Resettable
from ..exceptions import TrainError
from .callbacks import (
    StepCallback,
    EpochCallback,
    EpochPrinter,
    TrainCallback,
    TrainPrinter
)
from .checkpoints import Checkpoint, InMemory
from .schedulers import NoSchedule
from .data import TestDataBase, TrainDataBase

__all__ = ['Trainer']


class Trainer(ArgRepr):
    """Train and, optionally, evaluate a model with early stopping.

    Parameters
    ----------
    loss: Module
        PyTorch module that accepts the output(s) of the model to train as
        first argument(s) and the target as last argument and that produces
        the (scalar) loss to minimize, i.e., reduction must be "mean" or "sum".
    optimizer: Curry[Optimizer], optional
        A curry of a preconfigured PyTorch Optimizer (or some other custom
        construct) that returns a fully configured PyTorch Optimizer when
        called with the ``model.parameters()`` of the model to train. Defaults
        to ``AdamW`` with its default parameters.
    batch_size: int, optional
        Size of the mini-batches to request from the training data.
        Defaults to 64.
    max_epochs: int, optional
        Maximum number of epochs to train for. Defaults to 100.
    scheduler: Curry[LRScheduler], optional
        A curry of a PyTorch learning-rate scheduler (or some other custom
        construct) that returns a fully configured learning-rate scheduler
        when called with a PyTorch optimizer. Defaults to never changing the
        learning rate at all.
    warmup: int, optional
        In the beginning of training, the learning-rate `scheduler` will be
        stepped after every optimizer step for this many times. Afterward, it
        will only be stepped at the end of each epoch. Defaults to 0, which
        results in no warmup and the learning-rate `scheduler` being stepped
        at the end of the first epoch for the first time.
    batch_step: bool, optional
        Whether to step the learning-rate `scheduler` after each batch or after
        each epoch once the `warmup` period is over. Default to ``False``.
    patience: int, optional
        If patience is not ``None`` and smaller than `max_epochs`, early
        stopping is active. A snapshot of the model's state is taken after
        each epoch that improved the loss below its last minimum. If no
        improvement occurs for `patience` epochs, model training is stopped
        (even if `max_epochs` has not been reached yet) and the model is reset
        to its best state.
    max_n: int, optional
        Maximum number of data points to take from the training data to
        evaluate the train loss after each epoch. Defaults to number of data
        points in the test set (if present) or the train set (if not).
    step_freq: int, optional
        For how many batches to accumulate gradients before taking an
        optimization step. Defaults to 1, which corresponds to no accumulation.
    clip_grad: float, optional
        Clip gradients such that their overall norm is capped by the given
        value. Defaults to 1.0
    checkpoint: Checkpoint, optional
        Whenever the train (or test) loss after an epoch is smaller than the
        loss after the last, a new snapshot of the model state is saved by
        calling the `save` method of the `checkpoint` instance. Defaults to
        ``InMemory``.
    show_progress: bool, optional
        Whether to provide visual feedback to the console while training an
        epoch in the form of a progress bar. Defaults to ``True``
    step_cbs: iterable of StepCallback, optional
        All per-step callbacks will be called every `cb_freq` batches with the
        train loss of the last batch and the current learning rate. Defaults
        to an empty tuple, which does nothing.
    cb_freq: int, optional
        Number of batches to skip before calling `step_cb` again.
        Defaults to 1.
    epoch_cbs: iterable of EpochCallback, optional
        All epoch callbacks will be called after each epoch with epoch, train
        loss, test loss, and current learning rate. Defaults to a single
        ``EpochPrinter``.
    train_cbs: iterable of TrainCallback, optional
        All train callbacks will be called after training finished with last
        epoch, epoch with the best loss, the best loss itself, whether
        `max_epochs` was exhausted, and with the training history in the form
        of a dictionary of lists with train loss, test loss and learning rate.
        Defaults to a single ``TrainPrinter``.

    Important
    ---------
    optimizer
        Because the (partial) optimizer will simply be completed with
        ``model.parameters()``, parameter groups are not supported.
    scheduler
        During `warmup`, the scheduler is called after each optimizer step and,
        afterward, at the end of each epoch. The scheduler thus needs to be
        aware of the `warmup` settings and act accordingly.
    step_freq
        If the ``reduction`` of the `loss` is "mean", the loss of each batch is
        divided by this number before performing the backward pass. Strictly
        speaking, each batch should thus have the exact same number of data
        points in this case. Furthermore, complications might arise when
        the model to train contains elements like ``BatchNorm``.

    See Also
    --------
    InMemory
    EpochCallback
    EpochPrinter
    TrainPrinter

    """

    def __init__(
            self,
            loss: Module,
            optimizer: Curry[Optimizer] = Curry[AdamW](AdamW),
            batch_size: int = 64,
            max_epochs: int = 100,
            scheduler: Curry[LRScheduler] = Curry[NoSchedule](NoSchedule),
            warmup: int = 0,
            batch_step: bool = False,
            patience: int | None = None,
            max_n: int | None = None,
            step_freq: int = 1,
            clip_grad: float = 1.0,
            checkpoint: Checkpoint = InMemory(),
            show_progress: bool = True,
            step_cbs: Iterable[StepCallback] = (),
            cb_freq: int = 1,
            epoch_cbs: Iterable[EpochCallback] = EpochPrinter(),
            train_cbs: Iterable[TrainCallback] = TrainPrinter()
    ) -> None:
        self.loss = self.__sane(loss)
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.max_epochs = max_epochs
        self.scheduler = scheduler
        self.warmup = warmup
        self.batch_step = batch_step
        self.patience = max_epochs if patience is None else patience
        self.max_n = max_n
        self.step_freq = step_freq
        self.clip_grad = clip_grad
        self.checkpoint = checkpoint
        self.show_progress = show_progress
        self.step_cbs = step_cbs
        self.cb_freq = cb_freq
        self.epoch_cbs = epoch_cbs
        self.train_cbs = train_cbs
        super().__init__(
            loss,
            optimizer,
            batch_size,
            max_epochs,
            scheduler,
            warmup,
            batch_step,
            patience,
            max_n,
            step_freq,
            clip_grad,
            checkpoint,
            show_progress,
            step_cbs,
            cb_freq,
            epoch_cbs,
            train_cbs
        )
        self.history = {'train_loss': [], 'test_loss': [], 'lr': []}

    @staticmethod
    def __sane(loss: Module) -> Module:
        """Check that the "reduction" of the loss is not "none"."""
        cls = loss.__class__.__name__
        if not hasattr(loss, 'reduction'):
            msg = f'Loss {cls} must have an attribute "reduction"!'
            raise TypeError(msg)
        if loss.reduction == 'none':
            msg = f'The "reduction" of the {cls} cannot be "none"!'
            raise ValueError(msg)
        return loss

    @property
    def scale(self) -> float:
        """Scaling factor for the accumulated loss before the backward pass."""
        return 1.0 / self.step_freq if self.loss.reduction == 'mean' else 1.0

    def train(
            self,
            model: Resettable,
            train: TrainDataBase,
            test: TestDataBase | None = None
    ) -> Module:
        """Train a fresh model from scratch, starting from a clean slate.

        Parameters
        ----------
        model: Resettable
            PyTorch Module to train. Must have a ``reset_parameters()`` method
            that can be called without any parameters to re-initialize all
            trainable model parameters and buffers.
        train: TrainDataBase
            Training data.
        test: TestDataBase, optional
            Hold-out data to compute test loss. If configured, early stopping
            will track the `test` loss if `test` is given and the `train` loss
            if it is not.

        Returns
        -------
        Module
            The trained Pytorch model.

        Raises
        ------
        TrainError
            If the `model` does not have a ``reset_parameters()`` method.

        Important
        ---------
        The `model` to train must always return a *tuple* of tensors, not just
        a single tensor. If it produces only one tensor, return it as a 1-tuple
        of tensors.

        Warnings
        --------
        Model training is (re-)started from scratch every time this method is
        called. The training history is erased, all internal model parameters
        are reset to a pristine initial state, and previously saved checkpoints
        are irrevocably deleted!

        """
        self.checkpoint.reset_parameters()
        self.history = {'train_loss': [], 'test_loss': [], 'lr': []}
        try:
            model.reset_parameters()
        except Exception as error:
            msg = 'Models must have a working "reset_parameters()" method!"'
            raise TrainError(msg) from error
        model.zero_grad(set_to_none=True)
        return self.resume(model, train, test)

    def resume(
            self,
            model: Resettable,
            train: TrainDataBase,
            test: TestDataBase | None = None
    ) -> Module:
        """Resume model training from the best epoch checkpointed so far.

        Parameters
        ----------
        model: Module
            PyTorch Module to train.
        train: TrainDataBase
            Training data.
        test: TestDataBase, optional
            Hold-out data to compute test loss. If configured, early stopping
            will track the `test` loss if `test` is given and the `train` loss
            if it is not.

        Returns
        -------
        Module
            The trained Pytorch model.

        Note
        ----
        This is safe to use even if you have never trained your model before,
        and you are starting from scratch.

        Important
        ---------
        The `model` to train must always return a *tuple* of tensors, not just
        a single tensor. If it produces only one tensor, return it as a 1-tuple
        of tensors.

        """
        # Initialize training cycle.
        optimizer = self.optimizer(model.parameters())
        scheduler = self.scheduler(optimizer)
        epoch, best_loss = self.checkpoint.load(model, optimizer, scheduler)

        # Initialize counting and accumulation variables.
        best_epoch = epoch
        n_wait = 1
        max_epochs_reached = False
        ema = None  # Exponential moving average of per-batch training loss

        model.train()
        self.loss.train()
        # Loop over epochs.
        for epoch in range(epoch + 1, self.max_epochs + 1):
            # Get an iterator over batches for one epoch of training data.
            n_batches, batches = train(self.batch_size, self.step_freq, epoch)
            # Initialize a progress bar to monitor training in real time.
            progress = tqdm(
                batches,
                desc='Train',
                total=n_batches,
                leave=False,
                disable=not self.show_progress
            )
            # Loop over batches for one epoch of training data
            for batch_index, (features, target) in enumerate(progress, 1):
                loss = self.loss(*model(*features), target)
                # Scale gradients for multi-batch accumulation if required.
                (self.scale * loss).backward()
                # Clip gradients if greater than specified maximum norm.
                norm = clip_grad_norm_(model.parameters(), self.clip_grad)
                # Report loss and norm to the tqdm progress bar for feedback.
                grad = 'CLIP' if norm > self.clip_grad else f'{norm:4.2f}'
                # Exponentially smoothen the reported per-batch loss
                ema = loss.item() if ema is None else 0.5 * (loss.item() + ema)
                progress.set_postfix(loss=f'{ema:4.2f}', grad=grad)
                # Step after accumulating gradients for step_freq batches.
                if batch_index % self.step_freq == 0:
                    optimizer.step()
                    optimizer.zero_grad(set_to_none=True)
                    # Step the scheduler during warmup and maybe also afterward
                    if scheduler.last_epoch < self.warmup or self.batch_step:
                        scheduler.step()
                # Call the step callback with current loss and learning rate
                if batch_index % self.cb_freq == 0:
                    for step_cb in self.step_cbs:
                        step_cb(ema, scheduler.get_last_lr()[0], norm.item())

            # How many data points to take for computing train (and test) loss.
            n = train.n if test is None else test.n
            max_n = n if self.max_n is None else min(self.max_n, n)
            n_batches = math.ceil(max_n / self.batch_size)

            # Evaluate model on training data ...
            n = 0
            ema = None
            train_loss = 0.0
            model.eval()
            self.loss.eval()
            with pt.inference_mode():
                batches = train.sample(self.batch_size, max_n)
                progress = tqdm(
                    batches,
                    desc='Eval (train)',
                    total=n_batches,
                    leave=False,
                    disable=not self.show_progress
                )
                for features, target in progress:
                    loss = self.loss(*model(*features), target).item()
                    ema = loss if ema is None else 0.5 * (loss + ema)
                    progress.set_postfix(loss=f'{ema:4.2f}')
                    if self.loss.reduction == 'mean':
                        n_new = target.size(0)
                        train_loss += n_new * (loss - train_loss) / (n + n_new)
                        n += n_new
                    else:
                        train_loss += loss

            # ... and, if present, on test data.
            if test is None:
                test_loss = float('nan')
            else:
                n = 0
                ema = None
                test_loss = 0.0
                with pt.inference_mode():
                    batches = test.sample(self.batch_size)
                    progress = tqdm(
                        batches,
                        desc='Eval (test)',
                        total=n_batches,
                        leave=False,
                        disable=not self.show_progress
                    )
                    for features, target in progress:
                        loss = self.loss(*model(*features), target).item()
                        ema = loss if ema is None else 0.5 * (loss + ema)
                        progress.set_postfix(loss=f'{ema:4.2f}')
                        if self.loss.reduction == 'mean':
                            n_new = target.size(0)
                            test_loss += n_new * (loss - test_loss)/(n + n_new)
                            n += n_new
                        else:
                            test_loss += loss

            # Append epoch metrics to training history.
            current_lr = scheduler.get_last_lr()[0]
            self.history['train_loss'].append(train_loss)
            self.history['test_loss'].append(test_loss)
            self.history['lr'].append(current_lr)

            # Call callback with epoch metrics and information.
            if test is None:
                sample = train.sample(self.batch_size, max_n)
            else:
                sample = test.sample(self.batch_size)
            for epoch_cb in self.epoch_cbs:
                epoch_cb(
                    epoch,
                    train_loss,
                    test_loss,
                    current_lr,
                    model,
                    sample
                )

            # After warmup, step the scheduler each epoch if not set otherwise.
            if scheduler.last_epoch >= self.warmup and not self.batch_step:
                scheduler.step()

            # Check if the loss improved within our patience.
            track_loss = train_loss if test is None else test_loss
            if track_loss < best_loss:
                best_loss = track_loss
                best_epoch = epoch
                n_wait = 1
                self.checkpoint.save(
                    best_epoch,
                    best_loss,
                    model,
                    optimizer,
                    scheduler
                )
            elif n_wait < self.patience:
                n_wait += 1
            else:
                self.checkpoint.load(model, optimizer, scheduler)
                break

        # Did we exhaust the maximum number of epochs?
        else:
            max_epochs_reached = True

        # Call callbacks on finished training.
        for step_cb in self.step_cbs:
            step_cb.close()
        for epoch_cb in self.epoch_cbs:
            epoch_cb.close()
        for train_cb in self.train_cbs:
            train_cb(
                epoch,
                best_epoch,
                best_loss,
                max_epochs_reached,
                self.history
            )

        # Finally, return the trained model.
        return model



"""Collection of convenience classes that do not fit any other category."""

from .repr import ArgRepr, IndentRepr
from .enums import NotFound, LiteralNotFound, Bears, LiteralBears
from .loggers import (
    StdLogger,
    FileLogger,
    DEFAULT_FMT,
    SHORT_FMT,
    PID_FMT,
    RAW_FMT
)

__all__ = [
    'ArgRepr',
    'IndentRepr',
    'NotFound',
    'LiteralNotFound',
    'Bears',
    'LiteralBears',
    'StdLogger',
    'FileLogger',
    'DEFAULT_FMT',
    'SHORT_FMT',
    'PID_FMT',
    'RAW_FMT'
]



from typing import Any
from io import BytesIO
from collections.abc import Callable
import pandas as pd
import polars as pl
from ...misc import ArgRepr, Bears, LiteralBears
from .s3 import S3


class S3Parquet2DataFrame[T](ArgRepr):
    """Download a single parquet file from S3 object storage.

    Type-annotate classes on instantiation with either a pandas or a polars
    dataframe so that static type checkers can infer the return type of
    the callable instances!

    Parameters
    ----------
    s3: S3
        An instance of a wrapped S3 client.
    bucket: str
        The name of the bucket to upload to.
    prefix: str, optional
        The prefix of the parquet file to download. Since it (or part of it)
        can also be provided later, when the callable instance is called, it
        is optional here. Defaults to an empty string.
    bear: str, optional
        Type of dataframe to return. Can be one of "pandas" or "polars". Use
        the ``Bears`` enum to avoid typos. Defaults to "pandas".
    get_kws: dict, optional
        Keyword arguments (in addition to `Bucket` and `Key`) to pass to the
        `get_object <https://boto3.amazonaws.com/v1/documentation/api/latest/
        reference/services/s3/client/get_object.html>`_ method of the client.
        Defaults to ``None``.
    **kwargs
        Additional keyword arguments are passed on to the top-level
        ``read_parquet`` function of either pandas or polars.

    See Also
    --------
    S3
    ~swak.misc.Bears

    """

    def __init__(
            self,
            s3: S3,
            bucket: str,
            prefix: str = '',
            bear: str | Bears | LiteralBears = Bears.PANDAS,
            get_kws: dict[str, Any] | None = None,
            **kwargs: Any
    ) -> None:
        self.s3 = s3
        self.bucket = self.__strip(bucket)
        self.prefix = self.__strip(prefix)
        self.bear = bear.strip().lower()
        self.get_kws = {} if get_kws is None else get_kws
        self.kwargs = kwargs
        super().__init__(
            self.s3,
            self.bucket,
            self.prefix,
            self.bear,
            get_kws=get_kws,
            **self.kwargs
        )

    @staticmethod
    def __strip(attr: str | None) -> str:
        """Strip leading and trailing whitespaces from string arguments."""
        return attr if attr is None else attr.strip(' /')

    @property
    def read_parquet(self) -> Callable[[BytesIO, ...], T]:
        """Top-level ``read_parquet`` function of either pandas or polars."""
        return {
            Bears.PANDAS: pd.read_parquet,
            Bears.POLARS: pl.read_parquet
        }[self.bear]

    def __call__(self, path: str = '') -> T:
        """Download a single parquet file from S3 object storage.

        Parameters
        ----------
        path:
            The path to the parquet file to load. If given here, it will
            be appended to the `prefix` given at instantiation time.
            Defaults to an empty string.

        Returns
        -------
        DataFrame
            A pandas or polars dataframe, depending on `bear`.

        """
        stripped = path.strip(' /')
        prepended = '/' + stripped if stripped else stripped
        key = self.prefix + prepended
        response = self.s3.client.get_object(
            Key=key,
            Bucket=self.bucket,
            **self.get_kws
        )
        with BytesIO(response.get('Body').read()) as buffer:
            df = self.read_parquet(buffer, **self.kwargs)
        return df



from typing import TypedDict, Any
from collections.abc import Iterator
from abc import ABC, abstractmethod
from collections.abc import Callable
from ...misc import ArgRepr
from ..types import Module, Batches

__all__ = [
    'StepCallback',
    'StepPrinter',
    'EpochCallback',
    'EpochPrinter',
    'History',
    'TrainCallback',
    'TrainPrinter'
]


class History(TypedDict):
    """Summary of metrics passed to the callback when training is finished."""
    train_loss: list[float]  #: List of losses evaluated on train data.
    test_loss: list[float | None]  #: List of losses evaluated on test data.
    lr: list[float]  #: List of learning rates.


class StepCallback(ABC):
    """Base class to inherit from when implementing custom step callbacks."""

    @abstractmethod
    def __call__(
            self,
            train_loss: float,
            learning_rate: float,
            gradient_norm: float
    ) -> None:
        """Called after processing a batch to print, log, or save information.

        Parameters
        ----------
        train_loss: float
            The training loss on the current batch.
        learning_rate: float
            The learning rate used in the current optimization step.
        gradient_norm: float
            Norm of the current gradients.

        """
        ...

    @abstractmethod
    def close(self) -> None:
        """Called at the end of training to be compatible with TensorBoard."""
        ...


class EpochCallback(ABC):
    """Base class to inherit from when implementing custom epoch callbacks."""

    @abstractmethod
    def __call__(
            self,
            epoch: int,
            train_loss: int,
            test_loss: int,
            learning_rate: float,
            model: Module,
            data: Batches,
    ) -> None:
        """Called after each epoch to print, log, or otherwise analyse metrics.

        Parameters
        ----------
        epoch: int
            The current epoch in the training loop.
        train_loss: float
            The loss computed on a sample of the train data after the
            current `epoch`.
        test_loss: float
            The loss computed on a sample of the test data after the
            current `epoch`. Always ``nan`` if no test data is used.
        learning_rate: float
            Learning rate of the optimizer in the current `epoch`.
        model: Module
            A reference to the model being trained.
        data: Batches
            An iterator over 2-tuples of a feature-tensor tuple and a target
            tensor to call the `model` with. Sampled from the test data if
            present and from the train data otherwise.

        """
        ...

    @abstractmethod
    def close(self) -> None:
        """Called at the end of training to be compatible with TensorBoard."""
        ...


class TrainCallback(ABC):
    """Base class to inherit from when implementing custom train callbacks."""

    @abstractmethod
    def __call__(
            self,
            epoch: int,
            best_epoch: int,
            best_loss: float,
            max_epochs_reached: bool,
            history: History
    ) -> None:
        """Called after training has finished to print, log, or save a summary.

        Parameters
        ----------
        epoch: int
            The last epoch in the training loop.
        best_epoch: int
            The epoch with the lowest loss encountered.
        best_loss: float
            The lowest loss encountered.
        max_epochs_reached: bool
            Whether the maximum number of epochs was exhausted or not.
        history: History
            Dictionary with lists of train losses, test losses, and learning
            rates.

        """
        ...


class StepPrinter(ArgRepr, StepCallback):
    """Step callback assembling a one-liner on per-batch training progress.

    Parameters
    ----------
    printer: callable, optional
        Will be called with the assembled message. Defaults to the python
        builtin ``print`` function, but could also be a logging command.
    sep: str, optional
        The items concatenated into a one-line message will be separated by
        this string. Default to " ".

    """

    def __init__(
            self,
            printer: Callable[[str], Any] = print,
            sep: str = ', '
    ) -> None:
        super().__init__(printer, sep)
        self.printer = printer
        self.sep = sep

    def __iter__(self) -> Iterator['StepPrinter']:
        return iter([self])

    def __call__(
            self,
            train_loss: float,
            learning_rate: float,
            gradient_norm: float
    ) -> None:
        """Assemble one-liner with loss, lr, and norm, and call the printer."""
        msg = self.sep.join([
            f'{train_loss:7.5f}',
            f'{learning_rate:7.5f}',
            f'{gradient_norm:7.5f}'
        ])
        self.printer(msg)

    def close(self) -> None:
        """Does nothing because there is nothing to close."""


class EpochPrinter(ArgRepr, EpochCallback):
    """Epoch callback assembling an informative message on training progress.

    Parameters
    ----------
    printer: callable, optional
        Will be called with the assembled message. Defaults to the python
        builtin ``print`` function, but could also be a logging command.

    """

    def __init__(self, printer: Callable[[str], Any] = print) -> None:
        super().__init__(printer)
        self.printer = printer

    def __iter__(self) -> Iterator['EpochPrinter']:
        return iter([self])

    def __call__(
            self,
            epoch: int,
            train_loss: float,
            test_loss: float,
            learning_rate: float,
            model: Module,
            data: Batches
    ) -> None:
        """Assemble training-progress message and call the printer with it."""
        msg = (f'Epoch: {epoch:>3} | learning rate: {learning_rate:7.5f} | '
               f'train loss: {train_loss:7.5f} | test loss: {test_loss:7.5f}')
        self.printer(msg)

    def close(self) -> None:
        """Does nothing because there is nothing to close."""


class TrainPrinter(ArgRepr, TrainCallback):
    """Train callback assembling an informative message when training ends.

    Parameters
    ----------
    printer: callable, optional
        Will be called with the assembled message. Defaults to the python
        builtin ``print`` function, but could also be a logging command.

    """
    def __init__(self, printer: Callable[[str], Any] = print) -> None:
        super().__init__(printer)
        self.printer = printer

    def __iter__(self) -> Iterator['TrainPrinter']:
        return iter([self])

    def __call__(
            self,
            epoch: int,
            best_epoch: int,
            best_loss: float,
            max_epochs_reached: bool,
            history: History
    ) -> None:
        """Assemble a summary of model training and call the printer with it.

        Parameters
        ----------
        epoch: int
            The last epoch in the training loop.
        best_epoch: int
            The epoch with the lowest loss encountered.
        best_loss: float
            The lowest loss encountered.
        max_epochs_reached: bool
            Whether the maximum number of epochs was exhausted or not.
        history: History
            Dictionary with lists of train losses, test losses, and learning
            rates.

        """
        if max_epochs_reached:
            self.printer(f'Maximum number of {epoch} epochs exhausted!')
        else:
            msg = (f'Stopping after {epoch} epochs because, even after waiting'
                   f' for {epoch - best_epoch} more epochs, the loss did not '
                   f'drop below the lowest value of {best_loss:7.5f} seen in '
                   f'epoch {best_epoch}. Recovering that checkpoint.')
            self.printer(msg)



from typing import Any
from pandas import DataFrame
from pandas.core.groupby import DataFrameGroupBy
from ..misc import ArgRepr


class FrameGroupByAgg(ArgRepr):
    """Simple partial of for calling a grouped dataframe's ``agg`` method.

    Parameters
    ----------
    *args
        Arguments to pass on to the ``groupby`` method call.
    **kwargs
        Keyword arguments to pass on to the ``groupby`` method call.

    Notes
    -----
    See the pandas `agg docs <https://pandas.pydata.org/pandas-docs/
    stable/reference/api/pandas.core.groupby.DataFrameGroupBy.agg.html>`_ for
    a full list of (keyword) arguments and an extensive description of
    usage and configuration.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, grouped_df: DataFrameGroupBy) -> DataFrame:
        """Call a grouped dataframe`s ``agg`` method with the cached (kw)args.

        Parameters
        ----------
        grouped_df: DataFrameGroupBy
            The grouped dataframe to aggregate.

        Returns
        -------
        DataFrame
            The aggregation of the grouped dataframe.

        """
        return grouped_df.agg(*self.args, **self.kwargs)



import math
from typing import Any
from torch.nn.modules.module import _IncompatibleKeys
from ..types import Optimizer
from ...misc import ArgRepr

__all__ = [
    'NoSchedule',
    'LinearInverse',
    'LinearExponential',
    'LinearCosine'
]


class NoSchedule(ArgRepr):
    """Mock learning-rate scheduler that does not actually do anything.

    This class is just a placeholder for an actual PyTorch learning-rate
    scheduler. Any additional (keyword) arguments provided at instantiation
    are ignored.

    Parameters
    ----------
    optimizer: Optimizer
        A PyTorch optimizer instance.

    """

    def __init__(
            self,
            optimizer: Optimizer,
            *_: Any,
            **__: Any
    ) -> None:
        super().__init__(optimizer)
        self.optimizer = optimizer

    def step(self, *args: Any, **kwargs: Any) -> None:
        """Does not do anything, certainly not touching the learning rate.

        Positional and/or keyword arguments can be provided, but are ignored.

        """

    def get_last_lr(self) -> list[float]:
        """Returns the default learning rate of the optimizer."""
        return [self.optimizer.defaults['lr']]

    def state_dict(self) -> dict[str, Any]:
        """Because there is no state, return an empty dictionary."""
        return {}

    def load_state_dict(self, *_: Any, **__: Any) -> _IncompatibleKeys:
        """Do nothing, return the object thing as other methods of that type"""
        return _IncompatibleKeys([], [])


class LinearInverse(ArgRepr):
    """Scale up learning rate during warmup before decaying with inverse power.

    Instances of this class are not learning-rate schedulers by themselves!
    They are intended to be passed as ``lr_lambda`` argument to PyTorch's
    `LambdaLR <https://pytorch.org/docs/stable/generated/torch.optim.
    lr_scheduler.LambdaLR.html#torch.optim.lr_scheduler.LambdaLR>`__
    learning-rate scheduler.

    Parameters
    ----------
    warmup: int, optional
        Number of steps during which the learning rate will be linearly
        scaled up to the one specified in the optimizer. Defaults to 0,
        resulting in the learning rate already at its maximum value for the
        first step and only decaying thereafter.
    power: float, optional
        After `warmup` steps, the learning rate is scaled down with the
        inverse step number taken to the power of this number. Values are
        cut to lie in the interval [0.5, 1.0].
        Defaults to 0.5, the slowest decay.

    """

    def __init__(self, warmup: int = 0, power: float = 0.5) -> None:
        super().__init__(warmup, power)
        self.warmup = max(warmup, 0)
        self.power = min(max(power, 0.5), 1.0)

    def __call__(self, step: int) -> float:
        """Learning rate scaling factor depending on the step.

        Parameters
        ----------
        step: int
            Step to return the learning-rate scaling factor for.

        Returns
        -------
        float
            The learning-rate scaling factor

        """
        if step < self.warmup:
            return (step + 1) / self.warmup
        if step == self.warmup:
            return 1.0
        return (step - self.warmup) ** -self.power


class LinearExponential(ArgRepr):
    """Scale up learning rate during warmup before decaying exponentially.

    Instances of this class are not learning-rate schedulers by themselves!
    They are intended to be passed as ``lr_lambda`` argument to PyTorch's
    `LambdaLR <https://pytorch.org/docs/stable/generated/torch.optim.
    lr_scheduler.LambdaLR.html#torch.optim.lr_scheduler.LambdaLR>`__
    learning-rate scheduler.

    Parameters
    ----------
    warmup: int, optional
        Number of steps during which the learning rate will be linearly
        scaled up to the one specified in the optimizer. Defaults to 0,
        resulting in the learning rate already at its maximum value for the
        first step and only decaying thereafter.
    gamma: float, optional
        After `warmup` steps, the learning rate is scaled down with this
        number to the power of the step number. Therefore, it must lie in
        the interval (0.0, 1.0). Defaults to 0.95.

    """

    def __init__(self, warmup: int = 0, gamma: float = 0.95) -> None:
        super().__init__(warmup, gamma)
        self.warmup = max(warmup, 0)
        self.gamma = min(max(gamma, 0.0), 1.0)

    def __call__(self, step: int) -> float:
        """Learning rate scaling factor depending on the step.

        Parameters
        ----------
        step: int
            Step to return the learning-rate scaling factor for.

        Returns
        -------
        float
            The learning-rate scaling factor

        """
        if step < self.warmup:
            return (step + 1) / self.warmup
        return self.gamma ** (step - self.warmup)


class LinearCosine(ArgRepr):
    """Scale up learning rate during warmup before decaying with cosine.

    Instances of this class are not learning-rate schedulers by themselves!
    They are intended to be passed as ``lr_lambda`` argument to PyTorch's
    `LambdaLR <https://pytorch.org/docs/stable/generated/torch.optim.
    lr_scheduler.LambdaLR.html#torch.optim.lr_scheduler.LambdaLR>`__
    learning-rate scheduler.

    Parameters
    ----------
    warmup: int, optional
        Number of steps during which the learning rate will be linearly
        scaled up to the one specified in the optimizer. Defaults to 0,
        resulting in the learning rate already at its maximum value for the
        first step and only decaying thereafter.
    cooldown: int, optional
        After `warmup` steps, the learning rate is scaled down with a cosine
        function for this many steps. Defaults to 100, but must be at least 1.

    Notes
    -----
    The learning rate will never actually reach 0. Rather, it stays at the last
    value right before reaching 0. How small this value is, depends on the
    choice for `max_steps`. For `max_steps` = 1, it will stay at 1.0, for
    `max_steps` = 2, it will stay at 0.5, and so on.

    """

    def __init__(
            self,
            warmup: int = 0,
            cooldown: int = 100
    ) -> None:
        super().__init__(warmup, cooldown)
        self.warmup = max(warmup, 0)
        self.cooldown = max(1, cooldown)

    def __call__(self, step: int) -> float:
        """Learning rate scaling factor depending on the step.

        Parameters
        ----------
        step: int
            Step to return the learning-rate scaling factor for.

        Returns
        -------
        float
            The learning-rate scaling factor

        """
        if step < self.warmup:
            return (step + 1) / self.warmup
        if step < (self.warmup + self.cooldown):
            angle = math.pi * (step - self.warmup) / self.cooldown
            return 0.5 + 0.5 * math.cos(angle)
        angle = math.pi * (self.cooldown - 1) / self.cooldown
        return 0.5 + 0.5 * math.cos(angle)



class LossError(Exception):
    pass


class EmbeddingError(Exception):
    pass


class TrainError(Exception):
    pass


class CompileError(Exception):
    pass


class ValidationErrors(ExceptionGroup):
    pass


class ShapeError(Exception):
    pass


class DeviceError(Exception):
    pass


class DTypeError(Exception):
    pass



"""Special field types for type-annotating class variables in a JsonObject."""

from .maybe import Maybe
from .flexidate import FlexiDate
from .flexitime import FlexiTime
from .custom import CustomField
from .resolve import resolve
from .normalizers import Lower

__all__ = [
    'Maybe',
    'FlexiDate',
    'FlexiTime',
    'CustomField',
    'Lower',
    'resolve',
]
